
#include <fstream.h>

#include "proxy.h"
#include "strlib.h"

#define MAX_HEADER_LINE	16384
#define RELAY_BUF_SIZE	16384
#define MAX_METHOD	4
#define MAX_HOST	256
#define MAX_FILE	8192
#define MAX_PROTOCOL	16

void Proxy::fix_user_agent(char* relay_buf)
 {
  if(spoof_user_agent)
   if(!strncasecmp_to_delim(relay_buf, "User-agent", ':', RELAY_BUF_SIZE))
    {
     sprintf(relay_buf, "User-agent: %s\r\n", spoof_user_agent);
    }
 }

void Proxy::handle_client(int client_socket_fd)
 {
  char header_buf[MAX_HEADER_LINE];
  
  FILE *in  = fdopen(client_socket_fd, "r");
  FILE *out  = fdopen(client_socket_fd, "w");

  setbuf(in, NULL);
  setbuf(out, NULL);

  if(!in)
   {
    warn( "Proxy:fdopen failed on 'in': %s\n", strerror(errno));
    return;
   }

  if(!out)
   {
    warn( "fdopen failed on 'out': %s\n", strerror(errno));
    return;
   }

  char method[MAX_METHOD + 1];
  char host[MAX_HOST + 1];
  int port;
  char file[MAX_FILE + 1];
  char protocol[MAX_PROTOCOL + 1];
 
  int num_bytes;

  if(!fgets(header_buf, sizeof(header_buf), in))
   {
    warn("Error processing HTTP request");
    exit(1);
   }

  int header_buf_len = strlen(header_buf);
  int save_c = header_buf[header_buf_len - 1];
  header_buf[header_buf_len - 1]  = 0;
  info("%s %s", get_peer_addr(client_socket_fd), header_buf);
  header_buf[header_buf_len - 1]  = save_c;
 
  parse_request(header_buf, method, host, port, file, protocol);
  sprintf(header_buf, "%s %s %s\r\n", method, file, protocol);
  
  TcpClient* http_client = NULL;

    http_client = new TcpClient(host, port, log);
    http_client->connect();

   if(!*http_client)
    {
     warn("Aborted HTTP client");
     exit(0);
    }

   if((num_bytes = http_client->send(header_buf, strlen(header_buf))) == -1)
    {
     warn("Error sending the header to the server: %s\n", strerror(errno));
     return;
    }

  //log_error("Wrote %d bytes of the first line of the header\n",
  //  num_bytes);

  if(strcasecmp("GET", method) == 0)
   process_get(in, out, http_client);

  if(strcasecmp("POST", method) == 0)
   process_post(in, out, http_client);
 }

void Proxy::process_post(FILE* in, FILE* out, TcpClient* http_client)
 {
  int num_bytes;
  char relay_buf[RELAY_BUF_SIZE];
  int content_len = -1;

  while(1)
   {
      if(!fgets(relay_buf, sizeof(relay_buf), in))
       {
        warn("Unexpected termination of HTTP header");
	exit(0);
       }

      fix_user_agent(relay_buf);
       
      num_bytes = strlen(relay_buf);
      
      //fprintf(stderr, "num_bytes = %d\n", num_bytes);
      const char* content_len_str = "Content-length";
      char token_buf[sizeof(relay_buf)];

      int i;
      
      if(content_len < 0) 
       {
        token_buf[0] = 0;

	for(i = 0; i < sizeof(token_buf); i++)
         {
          if(isspace(relay_buf[i]) || relay_buf[i] == ':')
	   {
	    token_buf[i] = 0;
	    break;
      	   }
	  else
	    token_buf[i] = relay_buf[i];
	  }
	
	 if(strcasecmp(token_buf, content_len_str) == 0)
           { 
	     for(; i < sizeof(relay_buf); i++)
	       if(relay_buf[i] == ':') 
	         {
		  i++;
		  break;
		 }

 	      for(; i < sizeof(relay_buf); i++)
	       if(isdigit(relay_buf[i])) 
	         break;
 
	       content_len = 0;
	       //fprintf(stderr, "i = %d relay_buf +i = '%s'\n",
	       //  i, relay_buf + i);

 	     for(; i < sizeof(relay_buf); i++)
	      {
	       if(!isdigit(relay_buf[i])) 
	         break;
	       else
	        content_len = 10 * content_len + relay_buf[i] - '0'; 
	      }
           
	      //fprintf(stderr, "computed content_len = %d\n", content_len);
	    }
      	 }
       
     
      //fprintf(stderr, "final content_len = %d\n", content_len);
      if(relay_buf[0] == '\n' || relay_buf[0] == '\r' )
       {
        http_client->send("\r\n", 2);
        
	if(content_len > 0)
	 {
	  char* content_buf = new char[content_len];
	  while(content_len > 0)
	   {
	    num_bytes = fread(content_buf, 1, content_len, in);
	    char info_fmt_buf[64];
	    snprintf(info_fmt_buf, sizeof(info_fmt_buf), "POST content: %%%ds",
	      content_len);
	    info(info_fmt_buf, content_buf);
	    if(num_bytes <= 0)
	     {
	      warn("Error in POST request");
	      exit(1);
	     }
	    
	    http_client->send(content_buf, num_bytes);
	    content_len -= num_bytes;
	   }

	  delete[] content_buf;
	 }
	//fprintf(stderr, "End of header\n");
	break;
       }
    
      if( http_client->send(relay_buf, num_bytes) == -1)
        warn("Error sending the data: %s\n", strerror(errno));
   }

  while((num_bytes = http_client->receive(relay_buf, sizeof(relay_buf))) > 0)
   {
    //fprintf(stderr, "Writing to the proxy client:\n");
    //fwrite(relay_buf, 1, num_bytes, stderr);
    fwrite(relay_buf, 1, num_bytes, out);
   }
 }

void Proxy::process_get(FILE* in, FILE* out, TcpClient* http_client)
 {
  int num_bytes;
  char relay_buf[RELAY_BUF_SIZE];


  while(1)
   {
      if(!fgets(relay_buf, sizeof(relay_buf), in))
       {
        warn("Unexpected termination of HTTP header");
	exit(0);
       }
    
      fix_user_agent(relay_buf);
      num_bytes = strlen(relay_buf);

      if(relay_buf[0] == '\n' || relay_buf[0] == '\r' )
       {
        http_client->send("\r\n", 2);
        //fprintf(stderr, "End of header\n");
	break;
       }
    
      if( http_client->send(relay_buf, num_bytes) == -1)
        warn("Error sending the data: %s\n", strerror(errno));
   }

  while((num_bytes = http_client->receive(relay_buf, sizeof(relay_buf))) > 0)
   {
    //fprintf(stderr, "Writing to the proxy client:\n");
    //fwrite(relay_buf, 1, num_bytes, stderr);
    fwrite(relay_buf, 1, num_bytes, out);
   }
 }


void Proxy::parse_request(char *request, char* method, char* host,
   int& port,  char* file,
       char* protocol)
  {
   int j = 0;
   int i = 0;

   for(i = 0; request[j] != ' ' && i < MAX_METHOD; i++,j++)
    {
     method[i]  = request[j];
    }

   method[i] = 0;
   i =  0;

   for(; request[j] != '/' && j < MAX_HEADER_LINE; j++);
     
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }

   if(request[++j] != '/')
    {
     warn("request syntax error in request, '/' expected");
     exit(1);
    }

   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }

   j++;

   for(i = 0; i < MAX_HOST && request[j] != '/' && j < MAX_HEADER_LINE; i++,j++)
    {
     host[i] = request[j];
    }
 
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }
 
   if(i == MAX_HOST)
    {
     warn("host string too long");
     exit(1);
    }

   host[i] = 0;

   for(i = 0; i < MAX_FILE && request[j] != ' ' && j < MAX_HEADER_LINE; i++,j++)
    {
     file[i] = request[j];
    }
 
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }
 
   if(i == MAX_FILE)
    {
     warn("file string too long");
     exit(1);
    }

   file[i] = 0;
   j++;

   for(i = 0; i < MAX_PROTOCOL && 
     (request[j] != '\n' && request[j] != '\r') && j < MAX_HEADER_LINE; i++,j++)
    {
     protocol[i] = request[j];
    }
 
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }
 
   if(i == MAX_PROTOCOL)
    {
     warn("protocol string too long");
     exit(1);
    }

    protocol[i] = 0;

    char* str = strtok(host, ":");
    str = strtok(NULL, ":");
    if(str)
     {
      //fprintf(stderr, "str = %s\n", str);
      port = atoi(str);
      *(--str) = 0;
     }
    else
     port = 80;
  }

#ifdef DEBUG
int main(int c, char** args)
 {
    Proxy p(8000);
    int i;
    char* requests[] = {
     "GET http://www.direct1.com/cgi-bin/survey?hello=world HTTP/1.0\n",
     "GET http://www.faneuil.com:8000/cgi-bin/survey?hello=world HTTP/1.0\n",
       NULL};

    for(i = 0; requests[i]; i++)
     {
      char method[MAX_METHOD + 1];
      char host[MAX_HOST + 1];
      char file[MAX_FILE + 1];
      char protocol[MAX_PROTOCOL + 1];
      int port;

      //fprintf(stderr, "parsing %s\n", requests[i]);
      p.parse_request(requests[i], method, host, port, file, protocol);
      fprintf(stderr, "got method = '%s', host = '%s', port = %d, file = '%s', protocol = '%s'\n",
        method, host, port, file, protocol);
     }
 }
#endif



